extends CharacterBody2D

@export var center: Vector2 = Vector2(550, 370)  # Center of the orbit
@export var radius: float = 300.0                # Orbit radius
@export var orbit_speed: float = 0.5             # Angular speed (radians per second)
@export var dash_speed: float = 600.0            # Dash speed in pixels per second
@export var return_speed: float = 200.0          # Return speed in pixels per second

var angle: float = 0.0                          # Current orbit angle

# --- STATE FLAGS ---
var dashing: bool = false         # Currently dashing toward a penguin?
var returning: bool = false       # Returning back to orbit?
var charging: bool = false        # In charge/lock-on phase

# --- DASH VARIABLES ---
var dash_target: Vector2          # Final dash destination (locked penguin position)
var start_pos: Vector2            # Position at dash start
var return_target: Vector2        # Target point on orbit for returning

# --- LOCK/CHARGE VARIABLES ---
var lock_target_pos: Vector2 = Vector2.ZERO   # The locked penguin’s global position
var lock_direction: Vector2 = Vector2.ZERO    # Unit vector from orca to penguin
var lock_elapsed: float = 0.0                   # Elapsed time during charging (2 sec)

var charge_target: Vector2 = Vector2.ZERO       # Position on the orbit aligned with the penguin
var target_penguin: Penguin
var target_rotation: float = 0.0                # Rotation the orca must have to face the penguin

func _ready() -> void:
	# Start in orbit; after 8 seconds we’ll try to lock on.
	$LockTimer.start(8)

func _process(delta):
	if dashing:
		# --- DASHING STATE ---
		# Smoothly move toward the dash target.
		global_position = global_position.move_toward(dash_target, dash_speed * delta)
		var target_rot = (dash_target - global_position).angle() + deg_to_rad(90)
		rotation = lerp_angle(rotation, target_rot, delta * 5)
		
		if global_position.distance_to(dash_target) < 10.0:
			# On reaching the target, slice one ice floe using a unit line:
			dashing = false
			returning = true
			return_target = find_closest_point_on_radius()
			
	elif returning:
		# --- RETURNING STATE ---
		# Smoothly return to the orbit.
		global_position = global_position.move_toward(return_target, return_speed * delta)
		var target_rot = (return_target - global_position).angle() + deg_to_rad(90)
		rotation = lerp_angle(rotation, target_rot, delta * 5)
		
		if global_position.distance_to(return_target) < 10.0:
			returning = false
			$LockTimer.start(8)
			
	elif charging:
		# --- CHARGING (LOCK-ON) STATE ---
		# Stop movement while charging by not updating global_position.
		# Instead, you can still update the rotation if desired.
		rotation = lerp_angle(rotation, target_rotation, delta * 5)
		
		# Animate the dash line (a TextureRect child called "DashLine")
		lock_elapsed += delta
		# Over 2 seconds, scale the y–size from 0 to 1000
		var t = clamp(lock_elapsed / 2.0, 0, 1)
		$DashLine.size.y = t * 1000
		
		if lock_elapsed >= 2.0:
			# Just before dashing, reset the dash line (y size to 0), then dash.
			$DashLine.size.y = 0
			charging = false
			dash(lock_target_pos)
			
	else:
		# --- NORMAL ORBITING STATE ---
		angle += orbit_speed * delta
		var new_position = center + Vector2(cos(angle), sin(angle)) * radius
		# Use linear interpolation for smooth movement.
		global_position = global_position.lerp(new_position, 0.1)
		var target_rot = angle + deg_to_rad(180)
		rotation = lerp_angle(rotation, target_rot, delta * 5)

# Called to initiate the dash. The target is the locked penguin’s position.
func dash(target: Vector2) -> void:
	start_pos = global_position
	dash_target = target
	dashing = true
	# Make sure the dash line is hidden.
	$DashLine.size.y = 0
	
	await get_tree().create_timer(0.25).timeout
	slice_floes(start_pos, dash_target)

# Finds the closest point on the orbit (used when returning).
func find_closest_point_on_radius() -> Vector2:
	var current_angle = (global_position - center).angle()
	return center + Vector2(cos(current_angle), sin(current_angle)) * radius

# This timer callback starts the lock-on phase.
func _on_LockTimer_timeout() -> void:
	if Globals.current_state != Globals.GAME_STATE.INGAME:
		return
	# Get a random penguin from Globals.pengu_group.
	var pengu_group: Node2D = Globals.pengu_group
	var penguins = pengu_group.penguins
	if penguins.is_empty():
		return
	
	var tries = 0
	var random_index: int = randi() % penguins.size()
	while not is_instance_valid(penguins[random_index]) or penguins[random_index].is_queued_for_deletion():
		random_index = randi() % penguins.size()
		tries += 1
		if tries > 10:
			printerr("No valid penguins found. probably they are all dead but didnt sign off the pinguins group list")
			return
	var p = penguins[random_index]
	self.target_penguin = p
	
	# Lock onto the selected penguin.
	lock_target_pos = p.global_position
	lock_direction = (lock_target_pos - global_position).normalized()
	# Begin the 2–second charging phase.
	charging = true
	lock_elapsed = 0.0
	# Compute the charge target: the point on the orbit directly aligned with the penguin.
	var target_angle = (lock_target_pos - center).angle()
	charge_target = center + Vector2(cos(target_angle), sin(target_angle)) * radius
	# Set the target rotation so the orca faces the penguin.
	target_rotation = (lock_target_pos - global_position).angle() + deg_to_rad(90)

# Slices the first ice floe that the given unit–lengath line (with one end at the penguin) cuts.
func slice_floes(slice_start: Vector2, slice_end: Vector2) -> void:
	for child in Globals.ice_floes:
		#if child == target_penguin.get_parent(): # Only cut the slice with the target
		if not is_instance_valid(child) or child.is_queued_for_deletion():
			continue
		if child.will_line_cut_polygon(slice_start, slice_end):
			child.splice_with_line(slice_start, slice_end)
			break
