extends CharacterBody2D
class_name Player

@export var speed = 10.5
@export var friction: float = 1.015

# Grappling variables
var canPlace: bool = true
@export var FishScene: PackedScene
var Fish: Node = null

var CurrentEisscholle: Node2D
var lastGrabbedEisscholle: Node2D
var PositionOffset: Vector2
@export var IceScene: PackedScene
var Grappled: bool = false
@onready var rope_line: Line2D = $RopeLine

# Rubber band movement variables
var grapple_point: Vector2 = Vector2.ZERO
var grapple_influence: float = 0.6  # How strongly the grapple affects movement
var max_stretch_distance: float = 400.0  # Maximum distance before maximum tension
var min_tension_force: float = 0.5  # Minimum pull force when close
var max_tension_force: float = 5.0  # Maximum pull force at max stretch
var return_dampening: float = 0.92  # Reduces slingshot effect when controls are released

@export var area2D: Area2D
	
func _physics_process(delta):
	var movement_input = Vector2.ZERO
	
	# Only apply player input when not grappled or when player is actively trying to move
	if not Grappled or Input.is_action_pressed("MoveRight") or Input.is_action_pressed("MoveLeft") or Input.is_action_pressed("MoveUp") or Input.is_action_pressed("MoveDown"):
		# Get movement input
		var steering = Input.get_action_strength("MoveRight") - Input.get_action_strength("MoveLeft")
		var vertical = Input.get_action_strength("MoveUp") - Input.get_action_strength("MoveDown")
		movement_input = Vector2(steering * speed, -speed * vertical)
	
	# Apply movement input with dampening when grappled
	if Grappled:
		if movement_input != Vector2.ZERO:
			# Player is trying to move while grappled - apply reduced input
			velocity += movement_input * (1.0 - grapple_influence)
		
		# Apply the rubber band force
		apply_rubber_band_force(delta)
	else:
		# Normal movement when not grappled
		velocity += movement_input
	
	# Apply base friction
	velocity = velocity / friction
	
	if not $VisibleOnScreenNotifier2D.is_on_screen():
		var force = -Globals.level_center.global_position.direction_to(self.global_position) * 20
		velocity += force
	
	move_and_slide()

func apply_rubber_band_force(delta):
	if lastGrabbedEisscholle != null:
		# Calculate the grapple point (eisscholle position + offset)
		grapple_point = lastGrabbedEisscholle.position + PositionOffset
		
		# Get direction and distance to grapple point
		var to_grapple = grapple_point - position
		var distance = to_grapple.length()
		
		# Calculate tension force based on distance
		var tension_ratio = clamp(distance / max_stretch_distance, 0.0, 1.0)
		var tension_force = lerp(min_tension_force, max_tension_force, tension_ratio)
		
		# Apply rubber band force (stronger the further away)
		var pull_direction = to_grapple.normalized()
		var pull_force = pull_direction * tension_force * distance * 0.05
		
		# Apply pull force to player
		velocity += pull_force
		
		if distance > 10.0:  # Only pull object if reasonably far
			lastGrabbedEisscholle.apply_force((-pull_direction * tension_force * distance))

		# Apply counter-force when player tries to move away
		var movement_input = Vector2(
			Input.get_action_strength("MoveRight") - Input.get_action_strength("MoveLeft"),
			Input.get_action_strength("MoveUp") - Input.get_action_strength("MoveDown")
		)
		
		if movement_input != Vector2.ZERO:
			# Check if player is trying to move away from grapple point
			var moving_away = movement_input.normalized().dot(pull_direction) < -0.3
			
			if moving_away:
				# Apply increased resistance the further the player stretches
				var counter_force = -movement_input * tension_ratio * 0.5
				velocity += counter_force
		else:
			# When no input is given, dampen velocity to prevent slingshot effect
			velocity *= return_dampening

func _input(event: InputEvent) -> void:
	if Globals.current_state != Globals.GAME_STATE.INGAME:
		return
	
	if Input.is_action_pressed("Fish") and FishScene != null and canPlace:
		Fish = FishScene.instantiate()
		Fish.global_position = global_position	
		add_sibling(Fish)
		for pengu in Globals.pengu_group.penguins:
			if is_instance_valid(pengu) and not pengu.is_queued_for_deletion():
				pengu.set_target(Fish.global_position)
		canPlace = false
		$FishDespawnTimer.start(2)
		$FishCooldownTimer.start(4)	
		
	if Input.is_action_pressed("DebugGrapple") and IceScene != null and canPlace:
		var Ice = IceScene.instantiate()
		Ice.position = position	
		add_sibling(Ice)
		canPlace = false	
		$FishCooldownTimer.start(4)	
		print("Spawned")
		
	_grapple()	
	_ungrapple()

func _grapple() -> void:
	if Input.is_action_pressed("Grapple"):
		if CurrentEisscholle != null and not Grappled:
			print("Grabbing")
			lastGrabbedEisscholle = CurrentEisscholle
			PositionOffset = position - lastGrabbedEisscholle.position
			Grappled = true

func _process(delta: float) -> void:
	if Grappled and lastGrabbedEisscholle != null:
		var start_pos = to_local(position)
		var end_pos = to_local(lastGrabbedEisscholle.position + PositionOffset)

		# Number of segments for the rope
		var num_segments = 25
		var points = []

		# Calculate the distance for wave effect
		var distance = (end_pos - start_pos).length()
		var tension_ratio = clamp(distance / max_stretch_distance, 0.1, 1.0)
		
		# Calculate wave properties based on tension
		var wave_speed = 0.00005 * (1.0 / tension_ratio*20)  # Faster waves when looser	
		var wave_amplitude = 1.75 * (1.0 - tension_ratio * 0.9)  # Higher amplitude when looser

		# Create intermediate points along the rope
		for i in range(num_segments):
			var t = i / float(num_segments - 1)  # t goes from 0 to 1
			var segment_pos = start_pos.lerp(end_pos, t)  # Interpolate between start and end

			# Add wave effect that changes with tension
			var wave_offset = sin(Time.get_ticks_msec() * wave_speed + t * 10) * wave_amplitude

			segment_pos += Vector2(0, wave_offset)
			points.append(segment_pos)

		# Set the points to the rope line
		rope_line.points = points
		rope_line.visible = true
	else:
		rope_line.visible = false
		
	handle_animation()

func _ungrapple() -> void:
	if Input.is_action_just_released("Grapple") and Grappled:
		print("Un-Grabbed")
		Grappled = false
		lastGrabbedEisscholle = null
		
		# Apply a slight dampening when releasing to further reduce slingshot effect
		velocity *= 0.85

func _on_fish_despawn_timer_timeout() -> void:
	if Fish != null:
		Fish.queue_free()

func _on_fish_cooldown_timer_timeout() -> void:
	canPlace = true

func _on_area_2d_body_entered(body: Node2D) -> void:
	CurrentEisscholle = body


func _on_area_2d_body_exited(body: Node2D) -> void:
	CurrentEisscholle = null


func handle_animation() -> void:
	var input_strengths = [
		Input.get_action_strength("MoveLeft"),
		Input.get_action_strength("MoveUp"),
		Input.get_action_strength("MoveRight"),
		Input.get_action_strength("MoveDown")
	]
	var animations = ["left", "up", "right", "down"]
	var max_value = input_strengths.max()
	
	if max_value == 0:
		return
		
	var dir = input_strengths.find(max_value)
	$AnimationPlayer.play(animations[dir])
