extends RigidBody2D
class_name IceFloe

@export var floe_texture: Texture2D
@export var drift_velocity: float = 50.0
@export var zigzag_amplitude: float = 2.0
@export var zigzag_segments: int = 2
@export var zigzag_randomness: float = 3

@export var min_area: float = 10000.0

@onready var polygon_node: Polygon2D = $Polygon2D
@onready var shadow_polygon_node: Polygon2D = $ShadowPolygon2D
@onready var col_polygon_node: CollisionPolygon2D = $CollisionPolygon2D

var random_force = Vector2.ZERO
@export
var noise: FastNoiseLite
var speed_factor: float = 0.4

func _ready() -> void:
	var poly = polygon_node.polygon
	
	polygon_node.texture = floe_texture
	col_polygon_node.polygon = poly
	shadow_polygon_node.polygon = poly
	noise = noise.duplicate(true)
	noise.seed = randi()
	
	Globals.ice_floes.append(self)
	
	
	check_for_drown(poly)

func _process(delta: float) -> void:
	var offset = Vector2(0, 10).rotated(polygon_node.rotation)
	shadow_polygon_node.global_position = polygon_node.global_position + offset

func check_for_drown(poly):
	$ShadowPolygon2D.material = $ShadowPolygon2D.material.duplicate(true)
	if polygon_area(poly) < min_area:
		drown()

var drowned = false

func drown():
	drowned = true
	$AnimationPlayer.play("drown")
	await get_tree().create_timer(0.3).timeout
	for child in get_children():
		if child is Penguin:
			Globals.create_skull(child.global_position)
			Globals.pengu_group.n_penguins-=1
			if(Globals.pengu_group.n_penguins<=0):
				Globals.current_state=Globals.GAME_STATE.GAMEOVER

@onready
var bomb = preload("res://sprengfass/sprengfass.tscn")

func _exit_tree() -> void:
	if drowned:
		var bomb_instance = bomb.instantiate()
		get_parent().add_child(bomb_instance)
		bomb_instance.global_position = global_position
		Globals.ice_floes.erase(self)

func _integrate_forces(state: PhysicsDirectBodyState2D) -> void:
	# Random floating
	var t = Time.get_ticks_msec() / 10.0
	var force = Vector2(noise.get_noise_2d(t, 0), noise.get_noise_2d(t, 1000))
	force = force - force/2
	
	# Apply
	state.linear_velocity += force * speed_factor

func will_line_cut_polygon(line_start: Vector2, line_end: Vector2) -> bool:
	var local_start = to_local(line_start)
	var local_end = to_local(line_end)
	var polygon = polygon_node.polygon
	if !does_line_segment_intersect_polygon(local_start, local_end, polygon):
		return false
	var extended_slices = split_polygon_with_slice_line(polygon, [local_start, local_end])
	return extended_slices.size() == 2

func does_line_segment_intersect_polygon(line_start: Vector2, line_end: Vector2, polygon: PackedVector2Array) -> bool:
	var line_rect = Rect2(line_start, Vector2.ZERO).expand(line_end)
	var poly_rect = Rect2()
	for point in polygon:
		poly_rect = poly_rect.expand(point)
	if !line_rect.intersects(poly_rect):
		return false
	var n = polygon.size()
	for i in range(n):
		var j = (i + 1) % n
		if get_segments_intersection(line_start, line_end, polygon[i], polygon[j]):
			return true
	return is_point_in_polygon(line_start, polygon) or is_point_in_polygon(line_end, polygon)

func is_point_in_polygon(point: Vector2, polygon: PackedVector2Array) -> bool:
	var n = polygon.size()
	var inside = false
	var j = n - 1
	for i in range(n):
		if ((polygon[i].y > point.y) != (polygon[j].y > point.y)) and (point.x < (polygon[j].x - polygon[i].x) * (point.y - polygon[i].y) / (polygon[j].y - polygon[i].y) + polygon[i].x):
			inside = !inside
		j = i
	return inside

func get_segments_intersection(a: Vector2, b: Vector2, c: Vector2, d: Vector2) -> bool:
	var r = b - a
	var s = d - c
	var denom = r.cross(s)
	if abs(denom) < 0.0001:
		return false
	var t = (c - a).cross(s) / denom
	var u = (c - a).cross(r) / denom
	return t >= 0 and t <= 1 and u >= 0 and u <= 1

func splice_with_line(splice_start: Vector2, splice_end: Vector2) -> void:
	var original_polygon = polygon_node.polygon.duplicate()
	var global_pos = global_position
	var local_start = to_local(splice_start)
	var local_end = to_local(splice_end)
	if !does_line_segment_intersect_polygon(local_start, local_end, original_polygon):
		return
	var slice_line = [local_start, local_end]
	var slices = split_polygon_with_slice_line(original_polygon, slice_line)
	if slices.size() != 2:
		return
	var slice_dir = (local_end - local_start).normalized()
	var drift_dir = Vector2(-slice_dir.y, slice_dir.x)
	var center1 = calculate_polygon_center(slices[0])
	var center2 = calculate_polygon_center(slices[1])
	if (center2 - center1).dot(drift_dir) < 0:
		drift_dir = -drift_dir
	
	var pengus: Array[Penguin] = []
	for child in get_children():
		if child is Penguin:
			pengus.append(child)
			child.suspend()
	
	var floe1: IceFloe = self.duplicate()
	var floe2: IceFloe = self.duplicate()
	floe1.global_position = global_pos
	floe2.global_position = global_pos
	floe1.get_node("Polygon2D").polygon = slices[0]
	floe2.get_node("Polygon2D").polygon = slices[1]
	floe1.get_node("Polygon2D").texture = floe_texture
	floe2.get_node("Polygon2D").texture = floe_texture
	floe1.get_node("CollisionPolygon2D").polygon = floe1.get_node("Polygon2D").polygon
	floe2.get_node("CollisionPolygon2D").polygon = floe2.get_node("Polygon2D").polygon
	
	floe1.apply_impulse(-drift_velocity * drift_dir)
	floe2.apply_impulse(drift_velocity * drift_dir)
	get_parent().add_child(floe1)
	get_parent().add_child(floe2)
	queue_free()
	
	for pengu in pengus:
		pengu.unsuspend()

func calculate_polygon_center(polygon: PackedVector2Array) -> Vector2:
	var center = Vector2.ZERO
	var area = 0.0
	var n = polygon.size()
	for i in range(n):
		var j = (i + 1) % n
		var cross = polygon[i].cross(polygon[j])
		area += cross
		center += (polygon[i] + polygon[j]) * cross
	area *= 0.5
	if abs(area) < 0.0001:
		for point in polygon:
			center += point
		return center / n
	return center / (6.0 * area)

func get_line_intersection(a: Vector2, b: Vector2, c: Vector2, d: Vector2) -> Vector2:
	var r = b - a
	var s = d - c
	var denom = r.cross(s)
	if abs(denom) < 0.0001:
		return Vector2.ZERO
	var t = (c - a).cross(s) / denom
	return a + r * t

func create_zigzag_edge(start: Vector2, end: Vector2, segments: int, amplitude: float, randomness: float) -> Array:
	var points = []
	var dir = (end - start).normalized()
	var perp = Vector2(-dir.y, dir.x)
	for i in range(segments + 1):
		var t = i / float(segments)
		var base_point = start.lerp(end, t)
		if i == 0 or i == segments:
			points.append(base_point)
		else:
			var sign = 1.0 if (i % 2 == 0) else -1.0
			var random_factor = randf_range(1.0 - randomness, 1.0 + randomness)
			points.append(base_point + perp * amplitude * sign * random_factor)
	return points

func polygon_area(poly: Array) -> float:
	var area = 0.0
	var n = poly.size()
	for i in range(n):
		var j = (i + 1) % n
		area += poly[i].cross(poly[j])
	return abs(area) / 2.0

func clean_polygon(poly: Array) -> Array:
	if poly.is_empty():
		return poly
	var cleaned = [poly[0]]
	for i in range(1, poly.size()):
		if cleaned[cleaned.size() - 1].distance_to(poly[i]) > 0.01:
			cleaned.append(poly[i])
	if cleaned.size() > 1 and cleaned[0].distance_to(cleaned[cleaned.size() - 1]) < 0.01:
		cleaned.pop_back()
	return cleaned

func split_polygon_with_slice_line(polygon: PackedVector2Array, slice_line: Array) -> Array:
	var A = slice_line[0]
	var B = slice_line[1]
	var polygonBounds = Rect2()
	for point in polygon:
		polygonBounds = polygonBounds.expand(point)
	var sliceDir = (B - A).normalized()
	var diagonal = polygonBounds.size.length() * 2.0
	A = A - sliceDir * diagonal
	B = B + sliceDir * diagonal
	var side = []
	for p in polygon:
		side.append((B - A).cross(p - A))
	var intersections = []
	var indices = []
	for i in range(polygon.size()):
		var j = (i + 1) % polygon.size()
		if side[i] * side[j] <= 0:
			var inter = get_line_intersection(polygon[i], polygon[j], A, B)
			if inter != Vector2.ZERO:
				intersections.append(inter)
				indices.append(i)
	if intersections.size() != 2:
		return []
	if indices[0] > indices[1]:
		intersections.swap(0, 1)
		indices.swap(0, 1)
	var test_index = (indices[0] + 1) % polygon.size()
	var positive_side = side[test_index] >= 0
	var poly1 = []
	poly1.append(intersections[0])
	var idx = (indices[0] + 1) % polygon.size()
	while idx != (indices[1] + 1) % polygon.size():
		if (positive_side and side[idx] >= 0) or (not positive_side and side[idx] <= 0):
			poly1.append(polygon[idx])
		idx = (idx + 1) % polygon.size()
	poly1.append(intersections[1])
	var poly2 = []
	poly2.append(intersections[1])
	idx = (indices[1] + 1) % polygon.size()
	while idx != (indices[0] + 1) % polygon.size():
		if (positive_side and side[idx] < 0) or (not positive_side and side[idx] > 0):
			poly2.append(polygon[idx])
		idx = (idx + 1) % polygon.size()
	poly2.append(intersections[0])
	var line_length = intersections[0].distance_to(intersections[1])
	var dynamic_segments = max(2, int(zigzag_segments * line_length / 100.0))
	var zigzag_edge = create_zigzag_edge(intersections[0], intersections[1], dynamic_segments, zigzag_amplitude, zigzag_randomness)
	var zigzag_edge_reversed = zigzag_edge.duplicate()
	zigzag_edge_reversed.reverse()
	zigzag_edge_reversed.pop_front()
	for point in zigzag_edge_reversed:
		poly1.append(point)
	var zigzag_edge_normal = zigzag_edge.duplicate()
	zigzag_edge_normal.pop_front()
	for point in zigzag_edge_normal:
		poly2.append(point)
	poly1 = clean_polygon(poly1)
	poly2 = clean_polygon(poly2)
	if poly1.size() < 3 or poly2.size() < 3:
		return []
	if polygon_area(poly1) < 0.1 or polygon_area(poly2) < 0.1:
		return []
	return [poly1, poly2]
